/*
* Copyright (c) 2005, Graph Builder
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
* * Redistributions of source code must retain the above copyright notice,
* this list of conditions and the following disclaimer.
*
* * Redistributions in binary form must reproduce the above copyright notice,
* this list of conditions and the following disclaimer in the documentation
* and/or other materials provided with the distribution.
*
* * Neither the name of Graph Builder nor the names of its contributors may be
* used to endorse or promote products derived from this software without
* specific prior written permission.

* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

import javax.swing.*;
import java.awt.*;
import com.graphbuilder.curve.*;

public class CurveListCellRenderer extends DefaultListCellRenderer {

	public CurveListCellRenderer() {
		super();
	}

	public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
		String name = "<unknown>";

		if (value instanceof BezierCurve)
			name = "Bezier Curve";
		else if (value instanceof NURBSpline) // must come before BSpline
			name = "NURB-Spline";
		else if (value instanceof BSpline)
			name = "B-Spline";
		else if (value instanceof CardinalSpline)
			name = "Cardinal Spline";
		else if (value instanceof CatmullRomSpline)
			name = "Catmull-Rom Spline";
		else if (value instanceof CubicBSpline)
			name = "Cubic B-Spline";
		else if (value instanceof LagrangeCurve)
			name = "Lagrange Curve";
		else if (value instanceof NaturalCubicSpline)
			name = "Natural Cubic Spline";
		else if (value instanceof Polyline)
			name = "Polyline";

		return super.getListCellRendererComponent(list, name, index, isSelected, cellHasFocus);
	}
}
